/*
 * WinUsbIF.cpp: WinUSB Interface
 * Copyright: Keil - An ARM Company
 * Version: 1.0.0
 */

#include "stdafx.h"
#include "windows.h"

#include <stdio.h>

#undef _MP  // Required for WinDDK USB headers

extern "C"
{
#include "setupapi.h"
#include "winusb.h"
}

#include "WinUsbIF.h"

// Configuration defines
DEFINE_GUID( DEFAULT_WINUSB_DEVICE_INTERFACE_GUID, 0x7D9ADCFCL, 0xE570, 0x4B38, 0xBF, 0x4E, 0x8F, 0x81, 0xF6, 0x89, 0x64, 0xE0 );

// Local variables
static GUID           WinUsbDevIfGUID = DEFAULT_WINUSB_DEVICE_INTERFACE_GUID;   // Device Interface GUID
static int            WinUsbDevCount;                                           // Number of Devices
static WinUsbDevice_t WinUsbDevice [MAX_DEV_NUM];                               // Device runtime information

// Local functions
static void StrDesc2Str (PUSB_STRING_DESCRIPTOR strDscr, TCHAR *str);


// ---- User API functions -------------------------------------

/*
 *  WinUSB Interface: Initialization
 *    Parameters:      None
 *    Return Value:    None
 */
void WinUsbIF_Initialize (void) {

  WinUsbDevCount = 0;
  memset(WinUsbDevice, 0, sizeof(WinUsbDevice));
}


/*
 *  WinUSB Interface: Uninitialization
 *    Parameters:      None
 *    Return Value:    None
 */
void WinUsbIF_Uninitialize (void) {
  UCHAR dev;

  for (dev = 0; dev < MAX_DEV_NUM; dev++) {
    if (WinUsbDevice[dev].Open) { 
      WinUsbIF_CloseDevice(dev);
    }
  }
}


/*
 *  WinUSB Interface: Set Device Interface GUID
 *    Parameters:      guid:  Pointer to GUID
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_SetGUID (GUID *guid) {

  memcpy(&WinUsbDevIfGUID, guid, sizeof(GUID));

  return TRUE;
}


/*
 *  WinUSB Interface: Get Device Interface GUID
 *    Parameters:      guid:  Pointer to GUID
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_GetGUID (GUID *guid) {

  memcpy(guid, &WinUsbDevIfGUID, sizeof(GUID));

  return TRUE;
}


/*
 *  WinUSB Interface: Find Devices
 *    Parameters:      None
 *    Return Value:    Number of Devices found
 */
int WinUsbIF_FindDevices (void) {
  HDEVINFO                          deviceInfo;
  SP_DEVICE_INTERFACE_DATA          interfaceData;
  PSP_DEVICE_INTERFACE_DETAIL_DATA  detailData = NULL;
  TCHAR                            *ptr_tchar_start, *ptr_tchar_stop;
  ULONG                             length;
  UCHAR                             dev, mi;
  BOOL                              result, has_mi;

  // If there is an open Device close it first
  for (dev = 0; dev < MAX_DEV_NUM; dev++) {
    if (WinUsbDevice[dev].Open) { 
      WinUsbIF_CloseDevice(dev);
    }
  }

  // Initialize local variable and structure keeping Device information
  WinUsbDevCount  = 0;
  memset(WinUsbDevice, 0, sizeof(WinUsbDevice));

  // Get Device Information for all present devices
  deviceInfo = SetupDiGetClassDevs(&WinUsbDevIfGUID, 
                                    NULL, 
                                    NULL, 
                                    DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
  if (deviceInfo == INVALID_HANDLE_VALUE) { return 0; }

  interfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

  dev = -1;

  // Scan all Device
  do {
    dev++;

    // Device Interface Element of a Device Information set
    result = SetupDiEnumDeviceInterfaces( deviceInfo,
                                          NULL,
                                         &WinUsbDevIfGUID,
                                          dev,
                                         &interfaceData);
    if (result == FALSE) { break; }

    // Get Device Interface Details - Get Length
    SetupDiGetDeviceInterfaceDetail( deviceInfo,
                                    &interfaceData,
                                     NULL,
                                     0,
                                    &length,
                                     NULL);

    // Free / allocate memory for Device Detailed Data
    if (detailData != NULL) { free(detailData); }
    detailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(length);
    if (detailData == NULL) { break; }
      
    // Set cbSize in the detailData structure
    detailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

    // Get Device Interface Details
    result = SetupDiGetDeviceInterfaceDetail( deviceInfo,
                                             &interfaceData,
                                              detailData,
                                              length,
                                              NULL,
                                              NULL);
    if (result == FALSE) { continue; }

    // Check if we are processing Interface instance of Device with Multiple Interfaces
    if (_tcsstr(detailData->DevicePath, TEXT("mi_")) != NULL) {
      // Device with Multiple Interfaces
      has_mi = TRUE;

      // Find Device's currently processing MI (Multiple Interface)
      ptr_tchar_start = _tcsstr(detailData->DevicePath, TEXT("mi_"));
      ptr_tchar_stop  = _tcsstr(ptr_tchar_start,        TEXT("#"));

      mi = _ttoi(ptr_tchar_start + 3);                  // Current mi from string to number
      if ((mi != 0) && (WinUsbDevCount > 0)) {
        // Store Path to Device Interface
        _tcscpy_s(WinUsbDevice[WinUsbDevCount - 1].Interface[mi].DeviceInterfacePath, detailData->DevicePath);

        // Increase count of Interfaces that Device contains
        WinUsbDevice[WinUsbDevCount - 1].NumInterfaces++;
      }
    } else {
      // Device with Single Interface
      has_mi = FALSE;
      mi = 0;
    }

    if (mi == 0) {
      // If Multiple Interface index 0 or no Multiple Interfaces on Device

      // Store Path to Device Interface 0
      _tcscpy_s(WinUsbDevice[WinUsbDevCount].Interface[mi].DeviceInterfacePath, detailData->DevicePath);

        // Number of Interfaces that Device contains is 1 at the moment
      WinUsbDevice[WinUsbDevCount].NumInterfaces = 1;

      // Find Device's VID
      ptr_tchar_start = _tcsstr(detailData->DevicePath, TEXT("vid_"));
      ptr_tchar_stop  = _tcsstr(detailData->DevicePath, TEXT("&"));
    
      if ((ptr_tchar_start != NULL) && (ptr_tchar_stop != NULL)) {
        // Store Device's VID
        WinUsbDevice[WinUsbDevCount].VID = (USHORT)_tcstoul(ptr_tchar_start + 4, NULL, 16);
    
        // Find Device's PID
        ptr_tchar_start = _tcsstr(ptr_tchar_start, TEXT("pid_"));
        ptr_tchar_stop  = _tcsstr(ptr_tchar_start, TEXT("#"));
    
        if ((ptr_tchar_start != NULL) && (ptr_tchar_stop != NULL)) {
          // Store Device's PID
          WinUsbDevice[WinUsbDevCount].PID = (USHORT)_tcstoul(ptr_tchar_start + 4, NULL, 16);
    
          // Find Device's Serial Number or Unique ID
          ptr_tchar_start += 9;
          if (has_mi) {
            // Skip multiple interface index if it exists
            ptr_tchar_start = _tcsstr(ptr_tchar_start, TEXT("mi_")) + 6;
          }
          if (_tcsstr(ptr_tchar_start, TEXT("&")) != NULL) {
            // If Unique ID not Serial Number
            ptr_tchar_stop = _tcsstr(ptr_tchar_start, TEXT("&")) + 1;
            ptr_tchar_stop = _tcsstr(ptr_tchar_stop,  TEXT("&")) + 1;
            ptr_tchar_stop = _tcsstr(ptr_tchar_stop,  TEXT("&"));
          } else {
            // If Serial Number is available for Device
            ptr_tchar_stop = _tcsstr(ptr_tchar_start, TEXT("#"));
          }

          if ((ptr_tchar_start != NULL) && (ptr_tchar_stop != NULL)) {
            // Store Device's Serial Number or Unique ID
            _tcsncpy_s(WinUsbDevice[WinUsbDevCount].SerialNumber, (size_t)(ptr_tchar_stop - ptr_tchar_start + 1), ptr_tchar_start, _TRUNCATE);
          }
        }
      }
      WinUsbDevCount++;
    }
  } while (WinUsbDevCount < MAX_DEV_NUM);

  if (detailData != NULL) { free(detailData); }

  SetupDiDestroyDeviceInfoList(deviceInfo);

  return WinUsbDevCount;
}


/*
 *  WinUSB Interface: Get Device's number of Interfaces
 *    Parameters:      dev:    Device index
 *    Return Value:    Number of interfaces
 */
int  WinUsbIF_GetNumOfInterfaces (UCHAR dev) {

  if (dev >= MAX_DEV_NUM) { return 0; }

  return WinUsbDevice[dev].NumInterfaces;
}


/*
 *  WinUSB Interface: Get Device's Interface available Endpoints Mask
 *    Parameters:      dev:    Device index
 *                     intf:   Interface index
 *    Return Value:    32-bit mask representing available endpoints
 *                     (16-bits IN endpoints | 16-bits OUT endpoints)
 */
ULONG WinUsbIF_GetEndpointsMask (UCHAR dev, UCHAR intf) {

  if (dev  >= MAX_DEV_NUM)                     { return 0; }
  if (intf >= WinUsbDevice[dev].NumInterfaces) { return 0; }

  return WinUsbDevice[dev].Interface[intf].EpMask;
}


/*
 *  WinUSB Interface: Get Device's Name string (VID + PID + Serial Number or Unique ID)
 *    Parameters:      dev:     Device index
 *                     buf:     Pointer to buffer that receives string (depends on selected project character set)
 *                     max_len: Maximum length in number of characters that string can contain
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_GetName(UCHAR dev, TCHAR *buf, UCHAR max_len) {

  if (dev >= MAX_DEV_NUM)                                       { return FALSE; }
  if ((_tcslen(WinUsbDevice[dev].SerialNumber) + 20) > max_len) { return FALSE; }

  _stprintf_s(buf, max_len, TEXT("VID(%04X) PID(%04X) %s"), WinUsbDevice[dev].VID, WinUsbDevice[dev].PID, WinUsbDevice[dev].SerialNumber);

  return TRUE;
}


/*
 *  WinUSB Interface: Get Device's Manufacturer string
 *    Parameters:      dev:     Device index
 *                     buf:     Pointer to buffer that receives string (depends on selected project character set)
 *                     max_len: Maximum length in number of characters that string can contain
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_GetManufacturer(UCHAR dev, TCHAR *buf, UCHAR max_len) {

  if (dev >= MAX_DEV_NUM)                                { return FALSE; }
  if (_tcslen(WinUsbDevice[dev].Manufacturer) > max_len) { return FALSE; }

  _tcscpy(buf, WinUsbDevice[dev].Manufacturer);

  return TRUE;
}


/*
 *  WinUSB Interface: Get Device's Product string
 *    Parameters:      dev:     Device index
 *                     buf:     Pointer to buffer that receives string (depends on selected project character set)
 *                     max_len: Maximum length in number of characters that string can contain
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_GetProduct(UCHAR dev, TCHAR *buf, UCHAR max_len) {

  if (dev >= MAX_DEV_NUM)                           { return FALSE; }
  if (_tcslen(WinUsbDevice[dev].Product) > max_len) { return FALSE; }

  _tcscpy(buf, WinUsbDevice[dev].Product);

  return TRUE;
}


/*
*  WinUSB Interface: Get Device's Serial Number string
*    Parameters:      dev:     Device index
*                     buf:     Pointer to character array receiving Serial Number string (depends on selected project character set)
*                     max_len: Maximum length in number of characters that string can contain
*    Return Value:    TRUE - Success, FALSE - Error
*/
BOOL WinUsbIF_GetSerialNumber(UCHAR dev, TCHAR *buf, UCHAR max_len) {

    if (dev >= MAX_DEV_NUM)                                { return FALSE; }
    if (_tcslen(WinUsbDevice[dev].SerialNumber) > max_len) { return FALSE; }

    _tcscpy(buf, WinUsbDevice[WinUsbDevCount - 1].SerialNumber);

    return TRUE;
}


/*
*  WinUSB Interface: Get Device's Interface string
*    Parameters:      dev:     Device index
*                     intf:    Interface index
*                     buf:     Pointer to character array receiving Interface string (depends on selected project character set)
*                     max_len: Maximum length in number of characters that string can contain
*    Return Value:    TRUE - Success, FALSE - Error
*/
BOOL WinUsbIF_GetInterfaceString(UCHAR dev, UCHAR intf, TCHAR *buf, UCHAR max_len) {

    if (dev >= MAX_DEV_NUM)                                          { return FALSE; }
    if (intf >= WinUsbDevice[dev].NumInterfaces)                     { return FALSE; }
    if (_tcslen(WinUsbDevice[dev].Interface[intf].String) > max_len) { return FALSE; }

    _tcscpy(buf, WinUsbDevice[dev].Interface[intf].String);

    return TRUE;
}


/*
*  WinUSB Interface: Get Device's Vendor ID (VID)
*    Parameters:      dev:     Device index
*                     vid:     Pointer to variable receiving Vendor ID
*    Return Value:    TRUE - Success, FALSE - Error
*/
BOOL WinUsbIF_GetVID(UCHAR dev, WORD *vid) {

	if (dev >= MAX_DEV_NUM) { return FALSE; }

	*vid = WinUsbDevice[dev].VID;

	return TRUE;
}


/*
*  WinUSB Interface: Get Device's Product ID (PID)
*    Parameters:      dev:     Device index
*                     pid:     Pointer to variable receiving Product ID
*    Return Value:    TRUE - Success, FALSE - Error
*/
BOOL WinUsbIF_GetPID(UCHAR dev, WORD *pid) {

	if (dev >= MAX_DEV_NUM) { return FALSE; }

	*pid = WinUsbDevice[dev].PID;

	return TRUE;
}


/*
 *  WinUSB Interface: Open Device (and all containing Interfaces)
 *    Parameters:      dev:   Device index
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_OpenDevice (UCHAR dev) {
  HANDLE                   devIntfHandle;
  WINUSB_INTERFACE_HANDLE  winusbIntfHandle;
  USB_DEVICE_DESCRIPTOR    devDscr;
  USB_INTERFACE_DESCRIPTOR intfDscr;
  UCHAR                    strDscr[MAXIMUM_USB_STRING_LENGTH+2];
  WINUSB_PIPE_INFORMATION  pipeInfo;
  ULONG                    length;
  UCHAR                    intf_idx, ep_idx, ep_num, pipe_idx;
  BOOL                     result;

  if (dev >= MAX_DEV_NUM)                                         { return FALSE; }
  if (WinUsbDevice[dev].Interface[0].DeviceInterfacePath == NULL) { return FALSE; }

  // Go through all Interfaces
  for (intf_idx = 0; intf_idx < WinUsbDevice[dev].NumInterfaces; intf_idx++) {

    // Create File for Device Interface Read/Write
    devIntfHandle = CreateFile(WinUsbDevice[dev].Interface[intf_idx].DeviceInterfacePath, 
                               GENERIC_READ | GENERIC_WRITE, 
                               FILE_SHARE_READ | FILE_SHARE_WRITE, 
                               (LPSECURITY_ATTRIBUTES)NULL,
                               OPEN_EXISTING, 
                               FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, 
                               NULL);
  
    if (devIntfHandle == INVALID_HANDLE_VALUE) {
      return FALSE;
    }
  
    // Store Device Interface handle
    WinUsbDevice[dev].Interface[intf_idx].DeviceInterfaceHandle = devIntfHandle;
  
    // Initialize WinUSB for Device Interface and get WinUSB Device Interface handle
    result = WinUsb_Initialize(devIntfHandle, &winusbIntfHandle);
  
    if (intf_idx == 0) {
      // Get additional information from device (for Interface 0)

      if (result) {
        length = sizeof(UCHAR);

        // Get USB Bus speed and store it at the same time
        result = WinUsb_QueryDeviceInformation( winusbIntfHandle,
                                                DEVICE_SPEED,
                                               &length,
                                               &WinUsbDevice[dev].Speed);
      }
    
      if (result) {
        // Get Device Descriptor
        result = WinUsb_GetDescriptor( winusbIntfHandle,
                                       USB_DEVICE_DESCRIPTOR_TYPE,
                                       0, 0,
                                       (PUCHAR)&devDscr,
                                       sizeof(USB_DEVICE_DESCRIPTOR),
                                      &length);
      }
    
      if (result) {
        // Get Manufacturer string
        result = WinUsb_GetDescriptor( winusbIntfHandle,
                                       USB_STRING_DESCRIPTOR_TYPE,
                                       devDscr.iManufacturer, 0x0409,
                                       (PUCHAR)strDscr,
                                       MAXIMUM_USB_STRING_LENGTH+2,
                                      &length);
      }
    
      if (result) {
        // Store Manufacturer string
        StrDesc2Str((PUSB_STRING_DESCRIPTOR)strDscr, WinUsbDevice[dev].Manufacturer);

        // Get Product string
        result = WinUsb_GetDescriptor( winusbIntfHandle,
                                       USB_STRING_DESCRIPTOR_TYPE,
                                       devDscr.iProduct, 0x0409,
                                       (PUCHAR)strDscr,
                                       MAXIMUM_USB_STRING_LENGTH+2,
                                      &length);
      }
    
      if (result) {
        // Store Product string
        StrDesc2Str((PUSB_STRING_DESCRIPTOR)strDscr, WinUsbDevice[dev].Product);

        if (devDscr.iSerialNumber != 0) {
          // Get Serial Number string if it exists
          result = WinUsb_GetDescriptor(winusbIntfHandle,
                                        USB_STRING_DESCRIPTOR_TYPE,
                                        devDscr.iSerialNumber, 0x0409,
                                        (PUCHAR)strDscr,
                                        MAXIMUM_USB_STRING_LENGTH + 2,
                                       &length);
        }
      }

      // Clear stored Serial Number string
      memset(WinUsbDevice[dev].SerialNumber, 0, sizeof(WinUsbDevice[dev].SerialNumber));

      if (devDscr.iSerialNumber != 0) {
        if (result) {
          // Store Serial Number string if it exists
          StrDesc2Str((PUSB_STRING_DESCRIPTOR)strDscr, WinUsbDevice[dev].SerialNumber);
        }
      }
    }
    
    if (result) {
      // Store WinUSB Device Interface handle
      WinUsbDevice[dev].Interface[intf_idx].WinUsbInterfaceHandle = winusbIntfHandle;

      // Get information about Interface
      result = WinUsb_QueryInterfaceSettings( winusbIntfHandle,
                                              0,
                                             &intfDscr);
      if (result) {
        // Clear stored Interface string
        memset(WinUsbDevice[dev].Interface[intf_idx].String, 0, sizeof(WinUsbDevice[dev].Interface[intf_idx].String));

        if (intfDscr.iInterface != 0) {
          // Get Interface string if it exists
          result = WinUsb_GetDescriptor( winusbIntfHandle,
                                         USB_STRING_DESCRIPTOR_TYPE,
                                         intfDscr.iInterface, 0x0409,
                                         (PUCHAR)strDscr,
                                         MAXIMUM_USB_STRING_LENGTH+2,
                                        &length);
          if (result) {
            // Store Interface string if it exists
            StrDesc2Str((PUSB_STRING_DESCRIPTOR)strDscr, WinUsbDevice[dev].Interface[intf_idx].String);
          }
        }

        ep_num = intfDscr.bNumEndpoints;                // Get number of endpoints this Interface contains
        for (ep_idx = 0; ep_idx < ep_num; ep_idx++) {   // Go through all available endpoints of Interface
          // Get information about Pipe used for communication with endpoint
          result = WinUsb_QueryPipe( winusbIntfHandle,
                                     0,
                                     (UCHAR)ep_idx,
                                    &pipeInfo);
          if (result) {
            if (USB_ENDPOINT_DIRECTION_OUT(pipeInfo.PipeId)) {
              // Pipe for communicating with OUT endpoint
              // Increment number of OUT pipes, store PipeID and update mask of used endpoints
              pipe_idx = WinUsbDevice[dev].Interface[intf_idx].NumOutPipes++;
              WinUsbDevice[dev].Interface[intf_idx].OutPipeID[pipe_idx] = pipeInfo.PipeId;
              WinUsbDevice[dev].Interface[intf_idx].EpMask |= 1UL << pipeInfo.PipeId;
            } else {
              // Pipe for communicating with IN endpoint
              // Increment number of IN pipes, store PipeID and update mask of used endpoints
              pipe_idx = WinUsbDevice[dev].Interface[intf_idx].NumInPipes++;
              WinUsbDevice[dev].Interface[intf_idx].InPipeID [pipe_idx] = pipeInfo.PipeId;
              WinUsbDevice[dev].Interface[intf_idx].EpMask |= 1UL << 16 << (pipeInfo.PipeId & 0x0F);
            }
          }
        }
      }
    }
  }

  // Store opening status
  WinUsbDevice[dev].Open = result;

  return result;
}


/*
 *  WinUSB Interface: Close Device
 *    Parameters:      dev:   Device index
 *    Return Value:    None
 */
void WinUsbIF_CloseDevice (UCHAR dev) {
  UCHAR  intf_idx;
  size_t sz;

  if (dev >= MAX_DEV_NUM) { return; }

  if (WinUsbDevice[dev].Open) {
    for (intf_idx = 0; intf_idx < WinUsbDevice[dev].NumInterfaces; intf_idx++) {
      if (WinUsbDevice[dev].Interface[intf_idx].WinUsbInterfaceHandle != NULL) {
        // Close WinUSB Device Interface
        WinUsb_Free(WinUsbDevice[dev].Interface[intf_idx].WinUsbInterfaceHandle);
        WinUsbDevice[dev].Interface[intf_idx].WinUsbInterfaceHandle = NULL;
      }
      if (WinUsbDevice[dev].Interface[intf_idx].DeviceInterfaceHandle != NULL) {
        // Close Device Interface
        CloseHandle(WinUsbDevice[dev].Interface[intf_idx].DeviceInterfaceHandle);
        WinUsbDevice[dev].Interface[intf_idx].DeviceInterfaceHandle = NULL;
      }

      // Clear rest of interface structure instance (only path is left)
      sz =  sizeof(WinUsbInterface_t) - 
           (sizeof(WinUsbDevice[dev].Interface[intf_idx].DeviceInterfacePath)   + 
            sizeof(WinUsbDevice[dev].Interface[intf_idx].DeviceInterfaceHandle) + 
            sizeof(WinUsbDevice[dev].Interface[intf_idx].WinUsbInterfaceHandle)); 
      memset(&WinUsbDevice[dev].Interface[intf_idx].NumOutPipes, 0, sz);
    }

    // Clear Device information that is filled upon Device open
    memset(WinUsbDevice[dev].Manufacturer, 0, sizeof(WinUsbDevice[dev].Manufacturer));
    memset(WinUsbDevice[dev].Product,      0, sizeof(WinUsbDevice[dev].Product));
    WinUsbDevice[dev].Speed = 0;

    WinUsbDevice[dev].Open = FALSE;
  }
}


/*
 *  WinUSB Interface: Control Transfer
 *    Parameters:      dev:   Device index
 *                     sp:    Pointer to Setup Packet
 *                     buf:   Pointer to buffer containing data for stage data
 *                     sz:    Number of bytes to transfer in data stage
 *                     cnt:   Pointer to number of bytes transferred in data stage
 *                     ov:    Pointer to OVERLAPPED structure
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_ControlTransfer (UCHAR dev, WINUSB_SETUP_PACKET *sp, UCHAR *buf, ULONG sz, ULONG *cnt, OVERLAPPED *ov) {

  // WinUSB Control Transfer supports up to 4KB data stage transfer

  if (dev >= MAX_DEV_NUM)                                           { return FALSE; }
  if (sz  >  4096)                                                  { return FALSE; }
  if (sp  == NULL)                                                  { return FALSE; }
  if ((buf == NULL) && (sz != 0))                                   { return FALSE; }
  if (WinUsbDevice[dev].Interface[0].WinUsbInterfaceHandle == NULL) { return FALSE; }

  return (WinUsb_ControlTransfer( WinUsbDevice[dev].Interface[0].WinUsbInterfaceHandle,
                                 *sp,
                                  buf,
                                  sz,
                                  cnt,
                                  ov));
}


/*
 *  WinUSB Interface: Write Pipe (Bulk/Interrupt Endpoint)
 *    Parameters:      dev:   Device index
 *                     intf:  Interface index
 *                     pipe:  Pipe index
 *                     buf:   Pointer to buffer with data to write
 *                     sz:    Number of bytes to write
 *                     cnt:   Pointer to number of bytes written
 *                     ov:    Pointer to OVERLAPPED structure
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_WritePipe (UCHAR dev, UCHAR intf, UCHAR pipe, UCHAR *buf, ULONG sz, ULONG *cnt, OVERLAPPED *ov) {

  if (dev  >= MAX_DEV_NUM)                                             { return FALSE; }
  if (intf >= WinUsbDevice[dev].NumInterfaces)                         { return FALSE; }
  if ((buf == NULL) && (sz != 0))                                      { return FALSE; }
  if (WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle == NULL) { return FALSE; }
  if (WinUsbDevice[dev].Interface[intf].OutPipeID[pipe]       == NULL) { return FALSE; }

  return (WinUsb_WritePipe(WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle, 
                           WinUsbDevice[dev].Interface[intf].OutPipeID[pipe], 
                           buf, 
                           sz, 
                           cnt, 
                           ov));
}


/*
 *  WinUSB Interface: Read Pipe (Bulk/Interrupt Endpoint)
 *    Parameters:      dev:   Device index
 *                     intf:  Interface index
 *                     pipe:  Pipe index
 *                     buf:   Pointer to buffer that receives data
 *                     sz:    Number of bytes to read
 *                     cnt:   Pointer to number of bytes read
 *                     ov:    Pointer to OVERLAPPED structure
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_ReadPipe (UCHAR dev, UCHAR intf, UCHAR pipe, UCHAR *buf, ULONG sz, ULONG *cnt, OVERLAPPED *ov) {

  if (dev  >= MAX_DEV_NUM)                                             { return FALSE; }
  if (intf >= WinUsbDevice[dev].NumInterfaces)                         { return FALSE; }
  if ((buf == NULL) && (sz != 0))                                      { return FALSE; }
  if (WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle == NULL) { return FALSE; }
  if (WinUsbDevice[dev].Interface[intf].InPipeID[pipe]        == NULL) { return FALSE; }

  // Execute WinUSB call
  return (WinUsb_ReadPipe(WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle, 
                          WinUsbDevice[dev].Interface[intf].InPipeID[pipe], 
                          buf, 
                          sz, 
                          cnt, 
                          ov));
}


/*
 *  WinUSB Interface: Abort Pipe pending transfers (Bulk/Interrupt Endpoint)
 *    Parameters:      dev:      Device index
 *                     intf:     Interface index
 *                     pipe:     Pipe index
 *                     pipe_dir: Pipe direction (0 = OUT, 1 = IN)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_AbortPipe (UCHAR dev, UCHAR intf, UCHAR pipe, UCHAR pipe_dir) {
  UCHAR pipe_id;

  if (dev  >= MAX_DEV_NUM)                                             { return FALSE; }
  if (intf >= WinUsbDevice[dev].NumInterfaces)                         { return FALSE; }
  if (WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle == NULL) { return FALSE; }

  if (pipe_dir == 1) {
    pipe_id = WinUsbDevice[dev].Interface[intf].InPipeID[pipe];
  } else {
    pipe_id = WinUsbDevice[dev].Interface[intf].OutPipeID[pipe];
  }
  if (pipe_id == 0) { return FALSE; }

  return (WinUsb_AbortPipe(WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle, 
                           pipe_id));
}


/*
 *  WinUSB Interface: Flush Pipe cached data (Bulk/Interrupt Endpoint)
 *    Parameters:      dev:      Device index
 *                     intf:     Interface index
 *                     pipe:     Pipe index
 *                     pipe_dir: Pipe direction (0 = OUT, 1 = IN)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_FlushPipe (UCHAR dev, UCHAR intf, UCHAR pipe, UCHAR pipe_dir) {
  UCHAR pipe_id;

  if (dev  >= MAX_DEV_NUM)                                             { return FALSE; }
  if (intf >= WinUsbDevice[dev].NumInterfaces)                         { return FALSE; }
  if (WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle == NULL) { return FALSE; }

  if (pipe_dir == 1) {
    pipe_id = WinUsbDevice[dev].Interface[intf].InPipeID[pipe];
  } else {
    pipe_id = WinUsbDevice[dev].Interface[intf].OutPipeID[pipe];
  }
  if (pipe_id == 0) { return FALSE; }

  return (WinUsb_FlushPipe(WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle, 
                           pipe_id));
}


/*
 *  WinUSB Interface: Reset Pipe (Bulk/Interrupt Endpoint)
 *    Parameters:      dev:      Device index
 *                     intf:     Interface index
 *                     pipe:     Pipe index
 *                     pipe_dir: Pipe direction (0 = OUT, 1 = IN)
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_ResetPipe (UCHAR dev, UCHAR intf, UCHAR pipe, UCHAR pipe_dir) {
  UCHAR pipe_id;

  if (dev  >= MAX_DEV_NUM)                                             { return FALSE; }
  if (intf >= WinUsbDevice[dev].NumInterfaces)                         { return FALSE; }
  if (WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle == NULL) { return FALSE; }

  if (pipe_dir == 1) {
    pipe_id = WinUsbDevice[dev].Interface[intf].InPipeID[pipe];
  } else {
    pipe_id = WinUsbDevice[dev].Interface[intf].OutPipeID[pipe];
  }
  if (pipe_id == 0) { return FALSE; }

  // Execute WinUSB call
  return (WinUsb_ResetPipe(WinUsbDevice[dev].Interface[intf].WinUsbInterfaceHandle, 
                           pipe_id));
}


/*
 *  WinUSB Interface: Get Overlapped Result
 *    Parameters:      dev:   Device index
 *                     ov:    Pointer to OVERLAPPED structure
 *                     cnt:   Pointer to number of bytes transfered
 *                     wait:  Wait until operation completed
 *    Return Value:    TRUE - Success, FALSE - Error
 */
BOOL WinUsbIF_GetOverlappedResult (UCHAR dev, OVERLAPPED *ov, DWORD *cnt, BOOL wait) {

  if (dev >= MAX_DEV_NUM)                                           { return FALSE; }
  if (WinUsbDevice[dev].Interface[0].WinUsbInterfaceHandle == NULL) { return FALSE; }

  return (WinUsb_GetOverlappedResult(WinUsbDevice[dev].Interface[0].WinUsbInterfaceHandle, ov, cnt, wait));
}


// ---- Auxiliary Functions (local) ----------------------------

/*
 *  Convert String Descriptor to String
 *    Parameters:      str_desc: Pointer to String Descriptor
 *    Parameters:      str:      Pointer to converted String
 *    Return Value:    None
 */
static void StrDesc2Str (PUSB_STRING_DESCRIPTOR str_desc, TCHAR *str) {
  int n;

  n = 0;
  if ((str_desc->bDescriptorType == USB_STRING_DESCRIPTOR_TYPE) &&
      (str_desc->bLength > 2)                                   && 
      ((str_desc->bLength % 2) == 0)                            &&
      (str_desc->bString[0] != 0x0409)) {
    n = (str_desc->bLength - 2) / 2;
#if (TCHAR == WCHAR)
    wmemcpy_s(str, n, (TCHAR *)str_desc->bString, n);
#else
    int i;
    for (i = 0; i < n; i++) {
      str[i] = (UCHAR)str_desc->bString[i];
    }
#endif
  }
  str[n] = 0;                           // Terminate String
}
